"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const etag_1 = __importDefault(require("etag"));
const vary_1 = __importDefault(require("vary"));
const fresh_1 = __importDefault(require("fresh"));
const url_1 = require("url");
const mime_types_1 = __importDefault(require("mime-types"));
const destroy_1 = __importDefault(require("destroy"));
const path_1 = require("path");
const on_finished_1 = __importDefault(require("on-finished"));
const macroable_1 = require("macroable");
const utils_1 = require("@poppinss/utils");
const fs_1 = require("fs");
const content_disposition_1 = __importDefault(require("content-disposition"));
const cookie_1 = require("@poppinss/cookie");
function statFn(filePath) {
    return new Promise((resolve, reject) => {
        fs_1.stat(filePath, (error, stats) => {
            if (error) {
                reject(error);
                return;
            }
            resolve(stats);
        });
    });
}
class HttpException extends utils_1.Exception {
    static invoke(body, status) {
        if (body !== null && typeof (body) === 'object') {
            const error = new this(body.message || `Request aborted with status code${status}`, status);
            error.body = body;
            return error;
        }
        const error = new this(body || `Request aborted with status code${status}`, status);
        error.body = error.message;
        return error;
    }
    handle(error, ctx) {
        ctx.response.status(error.status).send(error.body);
    }
}
class Response extends macroable_1.Macroable {
    constructor(request, response, _config) {
        super();
        this.request = request;
        this.response = response;
        this._config = _config;
        this.lazyBody = null;
        this._headers = {};
        this._explicitStatus = false;
    }
    get hasLazyBody() {
        return !!(this.lazyBody && this.lazyBody.writer);
    }
    get finished() {
        return this.response.finished;
    }
    get headersSent() {
        return this.response.headersSent;
    }
    get isPending() {
        return (!this.headersSent && !this.finished);
    }
    _castHeaderValue(value) {
        return Array.isArray(value) ? value.map(String) : String(value);
    }
    _writeBody(content, generateEtag, jsonpCallbackName) {
        let { type, body, originalType } = this.buildResponseBody(content);
        if (body === null) {
            this.safeStatus(204);
            this.removeHeader('Content-Type');
            this.removeHeader('Content-Length');
            this.removeHeader('Transfer-Encoding');
            this._end();
            return;
        }
        if (type === 'unknown') {
            throw new Error(`Cannot send ${originalType} as HTTP response`);
        }
        if ([204, 304].indexOf(this.response.statusCode) > -1) {
            this.removeHeader('Content-Type');
            this.removeHeader('Content-Length');
            this.removeHeader('Transfer-Encoding');
            this._end(body);
            return;
        }
        if (jsonpCallbackName) {
            this.header('X-Content-Type-Options', 'nosniff');
            this.safeHeader('Content-Type', 'text/javascript; charset=utf-8');
        }
        else {
            this.safeHeader('Content-Type', `${type}; charset=utf-8`);
        }
        if (generateEtag) {
            this.setEtag(body);
        }
        if (jsonpCallbackName) {
            body = body.replace(/\u2028/g, '\\u2028').replace(/\u2029/g, '\\u2029');
            body = `/**/ typeof ${jsonpCallbackName} === 'function' && ${jsonpCallbackName}(${body});`;
        }
        this.header('Content-Length', Buffer.byteLength(body));
        this._end(body);
    }
    _streamBody(body, errorCallback) {
        return new Promise((resolve) => {
            let finished = false;
            body.on('error', (error) => {
                if (finished) {
                    return;
                }
                finished = true;
                destroy_1.default(body);
                if (typeof (errorCallback) === 'function') {
                    this._end(errorCallback(error));
                }
                else {
                    this._end(error.code === 'ENOENT' ? 'File not found' : 'Cannot process file', error.code === 'ENOENT' ? 404 : 500);
                    resolve();
                }
            });
            body.on('end', resolve);
            on_finished_1.default(this.response, () => {
                finished = true;
                destroy_1.default(body);
            });
            this.flushHeaders();
            body.pipe(this.response);
        });
    }
    async _download(filePath, generateEtag, errorCallback) {
        try {
            const stats = await statFn(filePath);
            if (!stats || !stats.isFile()) {
                throw new Error('response.download only accepts path to a file');
            }
            this.header('Last-Modified', stats.mtime.toUTCString());
            this.type(path_1.extname(filePath));
            if (generateEtag) {
                this.setEtag(stats, true);
            }
            if (this.request.method === 'HEAD') {
                this._end(null, generateEtag && this.fresh() ? 304 : 200);
                return;
            }
            if (generateEtag && this.fresh()) {
                this._end(null, 304);
                return;
            }
            this.header('Content-length', stats.size);
            return this._streamBody(fs_1.createReadStream(filePath), errorCallback);
        }
        catch (error) {
            if (typeof (errorCallback) === 'function') {
                this._end(errorCallback(error));
            }
            else {
                this._end('Cannot process file', 404);
            }
        }
    }
    _end(body, statusCode) {
        this.flushHeaders(statusCode);
        const res = this.response;
        res.end(body, null, null);
    }
    flushHeaders(statusCode) {
        this.response.writeHead(statusCode || this.response.statusCode, this._headers);
        this._headers = {};
        return this;
    }
    getHeader(key) {
        const value = this._headers[key.toLowerCase()];
        return value === undefined ? this.response.getHeader(key) : value;
    }
    header(key, value) {
        if (value) {
            this._headers[key.toLowerCase()] = this._castHeaderValue(value);
        }
        return this;
    }
    append(key, value) {
        if (!value) {
            return this;
        }
        key = key.toLowerCase();
        let existingHeader = this.getHeader(key);
        let casted = this._castHeaderValue(value);
        if (!existingHeader) {
            this._headers[key] = casted;
            return this;
        }
        existingHeader = this._castHeaderValue(existingHeader);
        casted = Array.isArray(existingHeader) ? existingHeader.concat(casted) : [existingHeader].concat(casted);
        this._headers[key] = casted;
        return this;
    }
    safeHeader(key, value) {
        if (!this.getHeader(key)) {
            this.header(key, value);
        }
        return this;
    }
    removeHeader(key) {
        delete this._headers[key.toLowerCase()];
        return this;
    }
    status(code) {
        this._explicitStatus = true;
        this.response.statusCode = code;
        return this;
    }
    safeStatus(code) {
        if (this._explicitStatus) {
            return this;
        }
        this.response.statusCode = code;
        return this;
    }
    type(type, charset) {
        type = charset ? `${type}; charset=${charset}` : type;
        this.header('Content-Type', mime_types_1.default.contentType(type));
        return this;
    }
    vary(field) {
        vary_1.default(this.response, field);
        return this;
    }
    setEtag(body, weak = false) {
        this.header('Etag', etag_1.default(body, { weak }));
        return this;
    }
    fresh() {
        if (this.request.method && ['GET', 'HEAD'].indexOf(this.request.method) === -1) {
            return false;
        }
        const status = this.response.statusCode;
        if ((status >= 200 && status < 300) || status === 304) {
            return fresh_1.default(this.request.headers, this._headers);
        }
        return false;
    }
    buildResponseBody(body) {
        if (body === null || body === undefined) {
            return {
                body: null,
                type: 'null',
            };
        }
        if (typeof (body) === 'string') {
            return body.length === 0
                ? {
                    body: null,
                    type: 'null',
                }
                : {
                    body,
                    type: /^\s*</.test(body) ? 'text/html' : 'text/plain',
                };
        }
        if (Buffer.isBuffer(body)) {
            return {
                body,
                type: 'application/octet-stream',
            };
        }
        if (typeof (body) === 'number' || typeof (body) === 'boolean') {
            return {
                body: String(body),
                type: 'text/plain',
            };
        }
        if (typeof (body) === 'object') {
            return {
                body: JSON.stringify(body),
                type: 'application/json',
            };
        }
        return {
            body,
            originalType: typeof (body),
            type: 'unknown',
        };
    }
    send(body, generateEtag = this._config.etag) {
        this.lazyBody = {
            writer: this._writeBody,
            args: [body, generateEtag],
        };
    }
    json(body, generateEtag) {
        return this.send(body, generateEtag);
    }
    jsonp(body, callbackName = this._config.jsonpCallbackName, generateEtag = this._config.etag) {
        this.lazyBody = {
            writer: this._writeBody,
            args: [body, generateEtag, callbackName],
        };
    }
    stream(body, errorCallback) {
        if (typeof (body.pipe) !== 'function' || !body.readable || typeof (body.read) !== 'function') {
            throw new Error('response.stream accepts a readable stream only');
        }
        this.lazyBody = {
            writer: this._streamBody,
            args: [body, errorCallback],
        };
    }
    download(filePath, generateEtag = this._config.etag, errorCallback) {
        this.lazyBody = {
            writer: this._download,
            args: [filePath, generateEtag, errorCallback],
        };
    }
    attachment(filePath, name, disposition, generateEtag, errorCallback) {
        name = name || filePath;
        this.header('Content-Disposition', content_disposition_1.default(name, { type: disposition }));
        return this.download(filePath, generateEtag, errorCallback);
    }
    location(url) {
        this.header('Location', url);
        return this;
    }
    redirect(url, sendQueryParams, statusCode = 302) {
        url = sendQueryParams ? `${url}?${url_1.parse(this.request.url, false).query}` : url;
        this.location(url);
        this.safeStatus(statusCode || 302);
        this.type('text/plain; charset=utf-8');
        this.send(`Redirecting to ${url}`);
    }
    abort(body, status) {
        throw HttpException.invoke(body, status || 400);
    }
    abortIf(condition, body, status) {
        if (condition) {
            this.abort(body, status);
        }
    }
    cookie(key, value, options) {
        options = Object.assign({}, this._config.cookie, options);
        const serialized = cookie_1.serialize(key, value, this._config.secret, options);
        if (!serialized) {
            return this;
        }
        this.append('set-cookie', serialized);
        return this;
    }
    plainCookie(key, value, options) {
        options = Object.assign({}, this._config.cookie, options);
        const serialized = cookie_1.serialize(key, value, undefined, options);
        if (!serialized) {
            return this;
        }
        this.append('set-cookie', serialized);
        return this;
    }
    clearCookie(key, options) {
        options = Object.assign({}, this._config.cookie, options);
        options.expires = new Date(1);
        const serialized = cookie_1.serialize(key, '', undefined, options);
        if (!serialized) {
            return this;
        }
        this.append('set-cookie', serialized);
        return this;
    }
    finish() {
        if (this.lazyBody && this.isPending) {
            this.lazyBody.writer.bind(this)(...this.lazyBody.args);
            this.lazyBody = null;
        }
        else if (this.isPending) {
            this._end();
        }
    }
}
exports.Response = Response;
Response._macros = {};
Response._getters = {};
